use std::hash::BuildHasherDefault;

use rustc_hash::FxHashMap;

use distribution_types::{LocalEditable, Requirements};
use pypi_types::Metadata23;
use uv_normalize::PackageName;

/// A set of editable packages, indexed by package name.
#[derive(Debug, Default, Clone)]
pub(crate) struct Editables(FxHashMap<PackageName, (LocalEditable, Metadata23, Requirements)>);

impl Editables {
    /// Create a new set of editables from a set of requirements.
    pub(crate) fn from_requirements(
        requirements: Vec<(LocalEditable, Metadata23, Requirements)>,
    ) -> Self {
        let mut editables =
            FxHashMap::with_capacity_and_hasher(requirements.len(), BuildHasherDefault::default());
        for (editable_requirement, metadata, requirements) in requirements {
            editables.insert(
                metadata.name.clone(),
                (editable_requirement, metadata, requirements),
            );
        }
        Self(editables)
    }

    /// Get the editable for a package.
    pub(crate) fn get(
        &self,
        name: &PackageName,
    ) -> Option<&(LocalEditable, Metadata23, Requirements)> {
        self.0.get(name)
    }

    /// Returns `true` if the given package is editable.
    pub(crate) fn contains(&self, name: &PackageName) -> bool {
        self.0.contains_key(name)
    }

    /// Iterate over all editables.
    pub(crate) fn iter(&self) -> impl Iterator<Item = &(LocalEditable, Metadata23, Requirements)> {
        self.0.values()
    }
}
